﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Interactivity;
using System.Windows.Media;

namespace Helpers {
   public class MouseOrganizeBehavior : Behavior<ListView> {

      protected static DependencyProperty isScrollEnabledProperty = DependencyProperty.Register("isScrollEnabled", typeof(bool), typeof(MouseOrganizeBehavior), new PropertyMetadata(true));


      public bool isScrollEnabled {
         get { return (bool)GetValue(isScrollEnabledProperty); }
         set { SetValue(isScrollEnabledProperty, value); }
      }


      public double UpperBound { get; set; }

      public double LowerBound { get; set; }


      private string dataType;

      public MouseOrganizeBehavior() {
         UpperBound = 10;
         LowerBound = 20;
      }

      protected override void OnAttached() {
         AssociatedObject.AllowDrop = true;


         dataType = String.Format("{0} - {1}", "ListViewOrderData", Guid.NewGuid().ToString());

         AssociatedObject.MouseMove += (s, e) => {
                if (e.LeftButton == System.Windows.Input.MouseButtonState.Pressed) {
                   if (AssociatedObject.SelectedItem == null) return;
                   DataObject d = new DataObject(dataType, AssociatedObject.SelectedItem);
                   DragDrop.DoDragDrop(AssociatedObject, d, DragDropEffects.Move | DragDropEffects.Scroll);
                }
             };
         AssociatedObject.DragOver += (s, e) => {
                e.Effects = DragDropEffects.None;
                if (e.Data.GetFormats().Contains(dataType)) {
                   var index = GetCurrentIndex(e.GetPosition);
                   MoveObject(DataToObj(e.Data), index);
                   ScrollList(index, e.GetPosition(AssociatedObject).Y);

                   e.Effects = DragDropEffects.Move | DragDropEffects.Scroll;
                }
                e.Handled = true;
             };
         AssociatedObject.Drop += (s, e) => {
                if (!e.Data.GetDataPresent(dataType)) return;
                object data = e.Data.GetData(dataType);

                var orgIndex = AssociatedObject.SelectedIndex;
                var index = GetCurrentIndex(e.GetPosition);
                if (index < 0) return;

                if (!MoveObject(data, index)) throw new Exception("Type not sortable!");
                e.Handled = true;
             };
      }

      protected override void OnDetaching() {
         AssociatedObject.AllowDrop = false;
         base.OnDetaching();
      }

      protected bool MoveObject(object obj, int targetIndex) {
         IList list = AssociatedObject.ItemsSource as IList ?? AssociatedObject.Items as IList;
         if (list == null) return false;
         if (targetIndex < 0) return false;

         list.Remove(obj);
         list.Insert(targetIndex, obj);
         return true;
      }

      protected void ScrollList(int targetIndex, double mouseY) {
         if (!isScrollEnabled || targetIndex < 0) return;
         if (mouseY < UpperBound) {
            if (targetIndex > 0)
               MoveListboxToIndex(targetIndex - 1);
         }
         else if (mouseY > AssociatedObject.ActualHeight - LowerBound)
            if (targetIndex <= AssociatedObject.Items.Count + 1 && !(targetIndex + 1 == AssociatedObject.Items.Count))
               MoveListboxToIndex(targetIndex + 1);

      }

      protected void MoveListboxToIndex(int index) {
         AssociatedObject.ScrollIntoView(AssociatedObject.Items[index]);
      }

      protected object DataToObj(IDataObject obj) {
         return obj.GetDataPresent(dataType) ? obj.GetData(dataType) : null;
      }

      protected int GetCurrentIndex(GetPositionDelegate getPosition) {
         int index = -1;
         for (int i = 0; i < AssociatedObject.Items.Count; ++i) {
            ListViewItem item = GetListViewItem(i);
            if (this.IsMouseOverTarget(item, getPosition)) {
               index = i;
               break;
            }
         }
         return index;
      }

      protected bool IsMouseOverTarget(Visual target, GetPositionDelegate getPosition) {
         if (target == null) return false;
         Rect bounds = VisualTreeHelper.GetDescendantBounds(target);
         Point mousePos = getPosition((IInputElement)target);
         return bounds.Contains(mousePos);
      }

      protected delegate Point GetPositionDelegate(IInputElement element);

      protected ListViewItem GetListViewItem(int index) {
         if (AssociatedObject.ItemContainerGenerator.Status != GeneratorStatus.ContainersGenerated)
            return null;

         return AssociatedObject.ItemContainerGenerator.ContainerFromIndex(index) as ListViewItem;
      }
   }
}
